@@ -12,47 +12,62 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-
+from __future__ import print_function, absolute_import
+from contextlib import contextmanager
 from argparse import ArgumentParser, ArgumentTypeError
 from fontmake import __version__
 from fontmake.font_project import FontProject
+from fontmake.errors import FontmakeError
+from ufo2ft.featureWriters import loadFeatureWriterFromString
 
 
-class PyClassType(object):
-    """ Callable object which returns a Python class defined in named module.
-    It can be passed as type= argument to ArgumentParser.add_argument().
-    """
-
-    def __init__(self, class_name):
-        self.class_name = class_name
-
-    def __call__(self, module_name):
-        import importlib
-        import inspect
-
+def _loadFeatureWriters(parser, specs):
+    feature_writers = []
+    for s in specs:
+        if s == 'None':
+            # magic value that means "don't generate any features!"
+            return []
         try:
-            mod = importlib.import_module(module_name)
-        except ImportError:
-            raise ArgumentTypeError("No module named %r" % module_name)
-
-        try:
-            klass = getattr(mod, self.class_name)
-        except AttributeError as e:
-            raise ArgumentTypeError("Module %r has no attribute %r"
-                                    % (module_name, self.class_name))
-
-        if not inspect.isclass(klass):
-            raise ArgumentTypeError("%r is not a class: %r"
-                                    % (self.class_name, type(klass)))
-        return klass
+            feature_writers.append(loadFeatureWriterFromString(s))
+        except Exception as e:
+            parser.error(
+                "Failed to load --feature-writer:\n  %s: %s"
+                 % (type(e).__name__, e)
+            )
+    return feature_writers
 
 
-def exclude_args(parser, args, excluded_args, source_name):
-    msg = '"%s" argument only available for %s source'
-    for excluded in excluded_args:
-        if args[excluded]:
-            parser.error(msg % (excluded, source_name))
-        del args[excluded]
+def exclude_args(parser, args, excluded_args, target):
+    msg = '"%s" option invalid for %s'
+    for argname in excluded_args:
+        if argname not in args:
+            continue
+        if args[argname]:
+            optname = "--%s" % argname.replace("_", "-")
+            parser.error(msg % (optname, target))
+        del args[argname]
+
+
+@contextmanager
+def _make_tempdirs(parser, args):
+    output = args["output"]
+    tempdirs = []
+    for dirname in ("master_dir", "instance_dir"):
+        if args.get(dirname) == "{tmp}":
+            if "ufo" in output:
+                parser.error(
+                    "Can't use temporary %s directory with 'ufo' output"
+                    % dirname.replace("_dir", ""))
+            import tempfile
+            td = args[dirname] = tempfile.mkdtemp(prefix=dirname+"_")
+            tempdirs.append(td)
+
+    yield tempdirs
+
+    if tempdirs:
+        import shutil
+        for td in tempdirs:
+            shutil.rmtree(td)
 
 
 def main(args=None):
@@ -78,9 +93,24 @@ def main(args=None):
         help='Output font formats. Choose between: %(choices)s. '
              'Default: otf, ttf',
         choices=('ufo', 'otf', 'ttf', 'ttf-interpolatable', 'variable'))
+    outputSubGroup = outputGroup.add_mutually_exclusive_group()
+    outputSubGroup.add_argument(
+        '--output-path', default=None,
+        help="Output font file path. Only valid when the output is a single "
+        "file (e.g. input is a single UFO or output is variable font)")
+    outputSubGroup.add_argument(
+        '--output-dir', default=None,
+        help="Output folder. By default, output folders are created in the "
+        "current working directory, grouping output fonts by format.")
     outputGroup.add_argument(
-        '-i', '--interpolate', action='store_true',
-        help='Interpolate masters (for Glyphs or MutatorMath sources only)')
+        '-i', '--interpolate', nargs="?", default=False, const=True,
+        metavar="INSTANCE_NAME",
+        help='Interpolate masters and generate all the instances defined. '
+             'To only interpolate a specific instance (or instances) that '
+             'match a given "name" attribute, you can pass as argument '
+             'the full instance name or a regular expression. '
+             'E.g.: -i "Noto Sans Bold"; or -i ".* UI Condensed". '
+             '(for Glyphs or MutatorMath sources only). ')
     outputGroup.add_argument(
         '-M', '--masters-as-instances', action='store_true',
         help='Output masters as instances')
@@ -90,6 +120,19 @@ def main(args=None):
     outputGroup.add_argument(
         '--round-instances', dest='round_instances', action='store_true',
         help='Apply integer rounding to all geometry when interpolating')
+    outputGroup.add_argument(
+        '--designspace-path', default=None,
+        help='Path to output designspace file (for Glyphs sources only).')
+    outputGroup.add_argument(
+        '--master-dir', default=None,
+        help='Directory where to write master UFO. Default: "./master_ufo". '
+             'If value is "{tmp}", a temporary directory is created and '
+             'removed at the end (for Glyphs sources only).')
+    outputGroup.add_argument(
+        '--instance-dir', default=None,
+        help='Directory where to write instance UFOs. Default: '
+             '"./instance_ufo". If value is "{tmp}", a temporary directory '
+             'is created and removed at the end (for Glyphs sources only).')
 
     contourGroup = parser.add_argument_group(title='Handling of contours')
     contourGroup.add_argument(
@@ -106,20 +149,31 @@ def main(args=None):
     contourGroup.add_argument(
         '-a', '--autohint', nargs='?', const='',
         help='Run ttfautohint. Can provide arguments, quoted')
+    contourGroup.add_argument(
+        '--cff-round-tolerance', type=float, default=None, metavar='FLOAT',
+        help='Restrict rounding of point coordinates in CFF table to only '
+             'those floats whose absolute difference from their integral part '
+             'is less than or equal to the tolerance. By default, all floats '
+             'are rounded to integer (tolerance 0.5); 0 disables rounding.'
+    )
 
     layoutGroup = parser.add_argument_group(title='Handling of OpenType Layout')
     layoutGroup.add_argument(
-        '--interpolate-binary-layout', action='store_true',
+        '--interpolate-binary-layout', nargs="?", default=False, const=True,
+        metavar="MASTER_DIR",
         help='Interpolate layout tables from compiled master binaries. '
              'Requires Glyphs or MutatorMath source.')
     layoutGroup.add_argument(
-        '--kern-writer-module', metavar="MODULE", dest='kern_writer_class',
-        type=PyClassType('KernFeatureWriter'),
-        help='Module containing a custom `KernFeatureWriter` class.')
-    layoutGroup.add_argument(
-        '--mark-writer-module', metavar="MODULE", dest='mark_writer_class',
-        type=PyClassType('MarkFeatureWriter'),
-        help='Module containing a custom `MarkFeatureWriter` class.')
+        "--feature-writer", metavar="CLASS", action="append",
+        dest="feature_writer_specs",
+        help="string specifying a feature writer class to load, either "
+             "built-in or from an external module, optionally initialized with "
+             "the given keyword arguments. The class and module names are "
+             "separated by '::'. The option can be repeated multiple times "
+             "for each writer class. A special value of 'None' will disable "
+             "all automatic feature generation. The option overrides both the "
+             "default ufo2ft writers and those specified in the UFO lib.")
+
     feaCompilerGroup = layoutGroup.add_mutually_exclusive_group(required=False)
     feaCompilerGroup.add_argument(
         '--use-afdko', action='store_true',
@@ -165,39 +219,54 @@ def main(args=None):
              '%(choices)s. Default: INFO')
     args = vars(parser.parse_args(args))
 
+    specs = args.pop("feature_writer_specs")
+    if specs is not None:
+        args["feature_writers"] = _loadFeatureWriters(parser, specs)
+
     glyphs_path = args.pop('glyphs_path')
     ufo_paths = args.pop('ufo_paths')
     designspace_path = args.pop('mm_designspace')
+    input_format = ("Glyphs" if glyphs_path else
+                    "designspace" if designspace_path else
+                    "UFO") + " source"
 
     if 'variable' in args['output']:
         if not (glyphs_path or designspace_path):
             parser.error(
                 'Glyphs or designspace source required for variable font')
-        for argname in ('interpolate', 'masters_as_instances',
-                        'interpolate_binary_layout'):
-            if args[argname]:
-                parser.error('--%s option invalid for variable font'
-                             % argname.replace("_", "-"))
+        exclude_args(parser, args,
+                     ['interpolate', 'masters_as_instances',
+                      'interpolate_binary_layout'],
+                     "variable output")
 
-    project = FontProject(timing=args.pop('timing'),
-                          verbose=args.pop('verbose'))
+    try:
+        project = FontProject(timing=args.pop('timing'),
+                              verbose=args.pop('verbose'))
 
-    if glyphs_path:
-        project.run_from_glyphs(glyphs_path, **args)
-        return
+        if glyphs_path:
+            with _make_tempdirs(parser, args):
+                project.run_from_glyphs(glyphs_path, **args)
+            return
 
-    exclude_args(parser, args, ['family_name', 'mti_source'], 'Glyphs')
-    if designspace_path:
-        project.run_from_designspace(designspace_path, **args)
-        return
+        exclude_args(parser, args,
+                     ['family_name', 'mti_source', 'designspace_path',
+                      'master_dir', 'instance_dir'],
+                     input_format)
+        if designspace_path:
+            project.run_from_designspace(designspace_path, **args)
+            return
 
-    exclude_args(
-        parser, args, ['interpolate', 'interpolate_binary_layout',
-                       'round_instances'],
-        'Glyphs or MutatorMath')
-    project.run_from_ufos(
-        ufo_paths, is_instance=args.pop('masters_as_instances'), **args)
+        exclude_args(parser, args,
+                     ['interpolate', 'interpolate_binary_layout',
+                      'round_instances'],
+                    input_format)
+        project.run_from_ufos(
+            ufo_paths, is_instance=args.pop('masters_as_instances'), **args)
+    except FontmakeError as e:
+        import sys
+        sys.exit("fontmake: error: %s" % e)
 
 
 if __name__ == '__main__':
-    main()
+    import sys
+    sys.exit(main())
